@@ -207,10 +207,10 @@ class IPPToolWrapper(object):
                     printer_uri=colander.null,
                     job_id=colander.null,
                     job_uri=colander.null):
-        kw = {'header': {'operation_attributes':
-                         {'printer_uri': printer_uri,
-                          'job_id': job_id,
-                          'job_uri': job_uri}}}
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'job_id': job_id,
+               'job_uri': job_uri}}
         request = release_job_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -221,11 +221,11 @@ class IPPToolWrapper(object):
                    job_id=colander.null,
                    job_uri=colander.null,
                    purge_job=colander.null):
-        kw = {'header': {'operation_attributes':
-                         {'printer_uri': printer_uri,
-                          'job_id': job_id,
-                          'job_uri': job_uri,
-                          'purge_job': purge_job}}}
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'job_id': job_id,
+               'job_uri': job_uri,
+               'purge_job': purge_job}}
         request = cancel_job_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -235,29 +235,45 @@ class IPPToolWrapper(object):
                    printer_uri=None,
                    job_name=colander.null,
                    ipp_attribute_fidelity=colander.null,
-                   document_name=colander.null,
-                   compression=colander.null,
-                   document_format=colander.null,
-                   document_natural_language=colander.null,
                    job_k_octets=colander.null,
                    job_impressions=colander.null,
                    job_media_sheets=colander.null,
+                   job_priority=colander.null,
+                   job_hold_until=colander.null,
+                   multiple_document_handling=colander.null,
+                   copies=colander.null,
+                   finishings=colander.null,
+                   page_ranges=colander.null,
+                   sides=colander.null,
+                   number_up=colander.null,
+                   orientation_requested=colander.null,
+                   printer_resolution=colander.null,
+                   print_quality=colander.null,
                    auth_info=colander.null,
                    job_billing=colander.null,
                    job_sheets=colander.null,
                    media=colander.null):
-        kw = {'header':
-              {'operation_attributes':
-               {'printer_uri': printer_uri,
-                'job_name': job_name,
-                'ipp_attribute_fidelity': ipp_attribute_fidelity,
-                'document_name': document_name,
-                'compression': compression,
-                'document_format': document_format,
-                'document_natural_language': document_natural_language,
-                'job_k_octets': job_k_octets,
-                'job_impressions': job_impressions,
-                'job_media_sheets': job_media_sheets}},
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'job_name': job_name,
+               'ipp_attribute_fidelity': ipp_attribute_fidelity,
+               'job_k_octets': job_k_octets,
+               'job_impressions': job_impressions,
+               'job_media_sheets': job_media_sheets},
+              'sub_operation_attributes':
+              {'job_priority': job_priority,
+               'job_hold_until': job_hold_until,
+               'job_sheets': job_sheets,
+               'multiple_document_handling': multiple_document_handling,
+               'copies': copies,
+               'finishings': finishings,
+               'page_ranges': page_ranges,
+               'sides': sides,
+               'number_up': number_up,
+               'orientation_requested': orientation_requested,
+               'media': media,
+               'printer_resolution': printer_resolution,
+               'print_quality': print_quality},
               'auth_info': auth_info,
               'job_billing': job_billing,
               'job_sheets': job_sheets,
@@ -278,24 +294,48 @@ class IPPToolWrapper(object):
                   job_k_octets=colander.null,
                   job_impressions=colander.null,
                   job_media_sheets=colander.null,
+                  job_priority=colander.null,
+                  job_hold_until=colander.null,
+                  multiple_document_handling=colander.null,
+                  copies=colander.null,
+                  finishings=colander.null,
+                  page_ranges=colander.null,
+                  sides=colander.null,
+                  number_up=colander.null,
+                  orientation_requested=colander.null,
+                  printer_resolution=colander.null,
+                  print_quality=colander.null,
                   auth_info=colander.null,
                   job_billing=colander.null,
                   job_sheets=colander.null,
                   media=colander.null,
                   document_content=None):
         filename, delete = _get_filename_for_content(document_content)
-        kw = {'header':
-              {'operation_attributes':
-               {'printer_uri': printer_uri,
-                'job_name': job_name,
-                'ipp_attribute_fidelity': ipp_attribute_fidelity,
-                'document_name': document_name,
-                'compression': compression,
-                'document_format': document_format,
-                'document_natural_language': document_natural_language,
-                'job_k_octets': job_k_octets,
-                'job_impressions': job_impressions,
-                'job_media_sheets': job_media_sheets}},
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'job_name': job_name,
+               'ipp_attribute_fidelity': ipp_attribute_fidelity,
+               'document_name': document_name,
+               'compression': compression,
+               'document_format': document_format,
+               'document_natural_language': document_natural_language,
+               'job_k_octets': job_k_octets,
+               'job_impressions': job_impressions,
+               'job_media_sheets': job_media_sheets},
+              'sub_operation_attributes':
+              {'job_priority': job_priority,
+               'job_hold_until': job_hold_until,
+               'job_sheets': job_sheets,
+               'multiple_document_handling': multiple_document_handling,
+               'copies': copies,
+               'finishings': finishings,
+               'page_ranges': page_ranges,
+               'sides': sides,
+               'number_up': number_up,
+               'orientation_requested': orientation_requested,
+               'media': media,
+               'printer_resolution': printer_resolution,
+               'print_quality': print_quality},
               'auth_info': auth_info,
               'job_billing': job_billing,
               'job_sheets': job_sheets,
@@ -312,9 +352,11 @@ class IPPToolWrapper(object):
     @pyipptool_coroutine
     def create_job_subscription(self,
                                 uri,
-                                printer_uri=None,
                                 requesting_user_name=None,
-                                notify_job_id=None,
+                                printer_uri=colander.null,
+                                job_id=colander.null,
+                                job_uri=colander.null,
+                                notify_job_id=colander.null,
                                 notify_recipient_uri=colander.null,
                                 notify_pull_method=colander.null,
                                 notify_events=colander.null,
@@ -324,11 +366,25 @@ class IPPToolWrapper(object):
                                 notify_time_interval=colander.null):
         """
         Create a per-job subscription object.
+
+        Create-Job-Subscription ???
+        http://www.cups.org/documentation.php/spec-ipp.html
+
+        OR
+
+        Create-Job-Subscriptions ???
+        https://tools.ietf.org/html/rfc3995#section-11.1.1
+
+        A ticket is open for the current bug in cups
+        https://www.cups.org/str.php?L4389
         """
-        kw = {'header': {'operation_attributes':
-                         {'printer_uri': printer_uri,
-                          'requesting_user_name': requesting_user_name,
-                          'notify_job_id': notify_job_id}},
+        raise NotImplementedError
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'requesting_user_name': requesting_user_name,
+               'notify_job_id': notify_job_id,
+               'job_id': job_id,
+               'job_uri': job_uri},
               'notify_recipient_uri': notify_recipient_uri,
               'notify_pull_method': notify_pull_method,
               'notify_events': notify_events,
@@ -357,9 +413,9 @@ class IPPToolWrapper(object):
         """
         Create a new subscription and return its id
         """
-        kw = {'header': {'operation_attributes':
-                         {'printer_uri': printer_uri,
-                          'requesting_user_name': requesting_user_name}},
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'requesting_user_name': requesting_user_name},
               'notify_recipient_uri': notify_recipient_uri,
               'notify_pull_method': notify_pull_method,
               'notify_events': notify_events,
@@ -390,8 +446,8 @@ class IPPToolWrapper(object):
                                 requesting_user_name_allowed=colander.null,
                                 requesting_user_name_denied=colander.null,
                                 printer_is_shared=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'printer_uri': printer_uri}},
+        kw = {'operation_attributes':
+             {'printer_uri': printer_uri},
               'auth_info_required': auth_info_required,
               'job_sheets_default': job_sheets_default,
               'device_uri': device_uri,
@@ -406,6 +462,7 @@ class IPPToolWrapper(object):
               'printer_state_message': printer_state_message,
               'requesting_user_name_allowed ': requesting_user_name_allowed,
               'requesting_user_name_denied': requesting_user_name_denied,
+              'printer_is_shared': printer_is_shared,
               }
 
         request = cups_add_modify_printer_form.render(kw)
@@ -427,8 +484,8 @@ class IPPToolWrapper(object):
                               requesting_user_name_allowed=colander.null,
                               requesting_user_name_denied=colander.null,
                               printer_is_shared=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'printer_uri': printer_uri}},
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri},
               'auth_info_required': auth_info_required,
               'member_uris': member_uris,
               'printer_is_accepting_jobs': printer_is_accepting_jobs,
@@ -440,6 +497,7 @@ class IPPToolWrapper(object):
               'printer_state_message': printer_state_message,
               'requesting_user_name_allowed ': requesting_user_name_allowed,
               'requesting_user_name_denied': requesting_user_name_denied,
+              'printer_is_shared': printer_is_shared,
               }
 
         request = cups_add_modify_class_form.render(kw)
@@ -448,14 +506,14 @@ class IPPToolWrapper(object):
 
     @pyipptool_coroutine
     def cups_delete_printer(self, uri, printer_uri=None):
-        kw = {'header': {'operation_attributes': {'printer_uri': printer_uri}}}
+        kw = {'operation_attributes': {'printer_uri': printer_uri}}
         request = cups_delete_printer_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
 
     @pyipptool_coroutine
     def cups_delete_class(self, uri, printer_uri=None):
-        kw = {'header': {'operation_attributes': {'printer_uri': printer_uri}}}
+        kw = {'operation_attributes': {'printer_uri': printer_uri}}
         request = cups_delete_class_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -469,14 +527,14 @@ class IPPToolWrapper(object):
                          printer_type_mask=colander.null,
                          requested_attributes=colander.null,
                          requested_user_name=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'first_printer_name': first_printer_name,
-                         'limit': limit,
-                         'printer_location': printer_location,
-                         'printer_type': printer_type,
-                         'printer_type_mask': printer_type_mask,
-                         'requested_attributes': requested_attributes,
-                         'requested_user_name': requested_user_name}}}
+        kw = {'operation_attributes':
+              {'first_printer_name': first_printer_name,
+               'limit': limit,
+               'printer_location': printer_location,
+               'printer_type': printer_type,
+               'printer_type_mask': printer_type_mask,
+               'requested_attributes': requested_attributes,
+               'requested_user_name': requested_user_name}}
         request = cups_get_classes_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -489,13 +547,13 @@ class IPPToolWrapper(object):
                          limit=colander.null,
                          requested_attributes=colander.null,
                          timeout=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'device_class': device_class,
-                         'exclude_schemes': exclude_schemes,
-                         'include-schemes': include_schemes,
-                         'limit': limit,
-                         'requested_attributes': requested_attributes,
-                         'timeout': timeout}}}
+        kw = {'operation_attributes':
+             {'device_class': device_class,
+              'exclude_schemes': exclude_schemes,
+              'include-schemes': include_schemes,
+              'limit': limit,
+              'requested_attributes': requested_attributes,
+              'timeout': timeout}}
         request = cups_get_devices_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -513,19 +571,19 @@ class IPPToolWrapper(object):
                       ppd_psversion=colander.null,
                       ppd_type=colander.null,
                       requested_attributes=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'exclude_schemes': exclude_schemes,
-                         'include_schemes': include_schemes,
-                         'limit': limit,
-                         'ppd_make': ppd_make,
-                         'ppd_make_and_model': ppd_make_and_model,
-                         'ppd_model_number': ppd_model_number,
-                         'ppd_natural_language': ppd_natural_language,
-                         'ppd_product': ppd_product,
-                         'ppd_psversion': ppd_psversion,
-                         'ppd_type': ppd_type,
-                         'requested_attributes': requested_attributes
-                         }}}
+        kw = {'operation_attributes':
+              {'exclude_schemes': exclude_schemes,
+               'include_schemes': include_schemes,
+               'limit': limit,
+               'ppd_make': ppd_make,
+               'ppd_make_and_model': ppd_make_and_model,
+               'ppd_model_number': ppd_model_number,
+               'ppd_natural_language': ppd_natural_language,
+               'ppd_product': ppd_product,
+               'ppd_psversion': ppd_psversion,
+               'ppd_type': ppd_type,
+               'requested_attributes': requested_attributes
+               }}
         request = cups_get_ppds_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -539,14 +597,14 @@ class IPPToolWrapper(object):
                           printer_type_mask=colander.null,
                           requested_attributes=colander.null,
                           requested_user_name=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'first_printer_name': first_printer_name,
-                         'limit': limit,
-                         'printer_location': printer_location,
-                         'printer_type': printer_type,
-                         'printer_type_mask': printer_type_mask,
-                         'requested_attributes': requested_attributes,
-                         'requested_user_name': requested_user_name}}}
+        kw = {'operation_attributes':
+              {'first_printer_name': first_printer_name,
+               'limit': limit,
+               'printer_location': printer_location,
+               'printer_type': printer_type,
+               'printer_type_mask': printer_type_mask,
+               'requested_attributes': requested_attributes,
+               'requested_user_name': requested_user_name}}
         request = cups_get_printers_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
@@ -557,11 +615,11 @@ class IPPToolWrapper(object):
                       job_id=colander.null,
                       job_uri=colander.null,
                       job_printer_uri=None,
-                      printer_state_message=None):
-        kw = {'header': {'operation_attributes':
-                        {'printer_uri': printer_uri,
-                         'job_id': job_id,
-                         'job_uri': job_uri}},
+                      printer_state_message=colander.null):
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'job_id': job_id,
+               'job_uri': job_uri},
               'job_printer_uri': job_printer_uri,
               'printer_state_message': printer_state_message}
         request = cups_move_job_form.render(kw)
@@ -573,9 +631,9 @@ class IPPToolWrapper(object):
                          printer_uri=None,
                          requesting_user_name=None,
                          printer_state_message=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'printer_uri': printer_uri,
-                         'requesting_user_name': requesting_user_name}},
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'requesting_user_name': requesting_user_name},
               'printer_state_message': printer_state_message}
         request = cups_reject_jobs_form.render(kw)
         response = yield self._call_ipptool(uri, request)
@@ -588,12 +646,12 @@ class IPPToolWrapper(object):
                            job_uri=colander.null,
                            requesting_user_name=colander.null,
                            requested_attributes=colander.null):
-        kw = {'header': {'operation_attributes':
-                        {'printer_uri': printer_uri,
-                         'job_id': job_id,
-                         'job_uri': job_uri,
-                         'requesting_user_name': requesting_user_name,
-                         'requested_attributes': requested_attributes}}}
+        kw = {'operation_attributes':
+              {'printer_uri': printer_uri,
+               'job_id': job_id,
+               'job_uri': job_uri,
+               'requesting_user_name': requesting_user_name,
+               'requested_attributes': requested_attributes}}
         request = get_job_attributes_form.render(kw)
         response = yield self._call_ipptool(uri, request)
         raise Return(response)
